@@ -37,6 +37,8 @@
 
 /* #define WRITE_SCHEDULER_DEBUG 1 */
 
+static const char *none_prefix[] = { NULL };
+
 static void g_at_chat_wakeup_writer(GAtChat *chat);
 static void debug_chat(GAtChat *chat, gboolean in, const char *str, gsize len);
 
@@ -87,6 +89,14 @@ struct _GAtChat {
 	guint wakeup_timeout;			/* How long to wait for resp */
 	GTimer *wakeup_timer;			/* Keep track of elapsed time */
 	GAtSyntax *syntax;
+	gboolean destroyed;			/* Re-entrancy guard */
+	GSList *terminator_list;		/* Non-standard terminator */
+};
+
+struct terminator_info {
+	char *terminator;
+	int len;
+	gboolean success;
 };
 
 static gint at_notify_node_compare_by_id(gconstpointer a, gconstpointer b)
@@ -137,7 +147,8 @@ static struct at_command *at_command_create(const char *cmd,
 						GAtNotifyFunc listing,
 						GAtResultFunc func,
 						gpointer user_data,
-						GDestroyNotify notify)
+						GDestroyNotify notify,
+						gboolean wakeup)
 {
 	struct at_command *c;
 	gsize len;
@@ -176,12 +187,16 @@ static struct at_command *at_command_create(const char *cmd,
 	/* If we have embedded '\r' then this is a command expecting a prompt
 	 * from the modem.  Embed Ctrl-Z at the very end automatically
 	 */
-	if (strchr(cmd, '\r'))
-		c->cmd[len] = 26;
-	else
-		c->cmd[len] = '\r';
+	if (wakeup == FALSE) {
+		if (strchr(cmd, '\r'))
+			c->cmd[len] = 26;
+		else
+			c->cmd[len] = '\r';
 
-	c->cmd[len+1] = '\0';
+		len += 1;
+	}
+
+	c->cmd[len] = '\0';
 
 	c->expect_pdu = expect_pdu;
 	c->prefixes = prefixes;
@@ -203,6 +218,14 @@ static void at_command_destroy(struct at_command *cmd)
 	g_free(cmd);
 }
 
+static void free_terminator(struct terminator_info *info)
+{
+	g_free(info->terminator);
+	info->terminator = NULL;
+	g_free(info);
+	info = NULL;
+}
+
 static void g_at_chat_cleanup(GAtChat *chat)
 {
 	struct at_command *c;
@@ -250,6 +273,13 @@ static void g_at_chat_cleanup(GAtChat *chat)
 	chat->syntax = NULL;
 
 	chat->channel = NULL;
+
+	if (chat->terminator_list) {
+		g_slist_foreach(chat->terminator_list,
+					(GFunc)free_terminator, NULL);
+		g_slist_free(chat->terminator_list);
+		chat->terminator_list = NULL;
+	}
 }
 
 static void read_watcher_destroy_notify(GAtChat *chat)
@@ -259,6 +289,9 @@ static void read_watcher_destroy_notify(GAtChat *chat)
 
 	if (chat->user_disconnect)
 		chat->user_disconnect(chat->user_disconnect_data);
+
+	if (chat->destroyed)
+		g_free(chat);
 }
 
 static void write_watcher_destroy_notify(GAtChat *chat)
@@ -355,12 +388,6 @@ static void g_at_chat_finish_command(GAtChat *p, gboolean ok,
 	at_command_destroy(cmd);
 }
 
-struct terminator_info {
-	const char *terminator;
-	int len;
-	gboolean success;
-};
-
 static struct terminator_info terminator_table[] = {
 	{ "OK", -1, TRUE },
 	{ "ERROR", -1, FALSE },
@@ -374,6 +401,27 @@ static struct terminator_info terminator_table[] = {
 	{ "+EXT ERROR:", 11, FALSE }
 };
 
+void g_at_chat_add_terminator(GAtChat *chat, char *terminator,
+					int len, gboolean success)
+{
+	struct terminator_info *info = g_new0(struct terminator_info, 1);
+	info->terminator = g_strdup(terminator);
+	info->len = len;
+	info->success = success;
+	chat->terminator_list = g_slist_prepend(chat->terminator_list, info);
+}
+
+static gboolean check_terminator(struct terminator_info *info, char *line)
+{
+	if (info->len == -1 && !strcmp(line, info->terminator))
+		return TRUE;
+
+	if (info->len > 0 && !strncmp(line, info->terminator, info->len))
+		return TRUE;
+
+	return FALSE;
+}
+
 static gboolean g_at_chat_handle_command_response(GAtChat *p,
 							struct at_command *cmd,
 							char *line)
@@ -381,17 +429,19 @@ static gboolean g_at_chat_handle_command_response(GAtChat *p,
 	int i;
 	int size = sizeof(terminator_table) / sizeof(struct terminator_info);
 	int hint;
+	GSList *l;
 
 	for (i = 0; i < size; i++) {
 		struct terminator_info *info = &terminator_table[i];
-
-		if (info->len == -1 && !strcmp(line, info->terminator)) {
+		if (check_terminator(info, line)) {
 			g_at_chat_finish_command(p, info->success, line);
 			return TRUE;
 		}
+	}
 
-		if (info->len > 0 &&
-			!strncmp(line, info->terminator, info->len)) {
+	for (l = p->terminator_list; l; l = l->next) {
+		struct terminator_info *info = l->data;
+		if (check_terminator(info, line)) {
 			g_at_chat_finish_command(p, info->success, line);
 			return TRUE;
 		}
@@ -587,7 +637,9 @@ static void new_bytes(GAtChat *p)
 
 	GAtSyntaxResult result;
 
-	while (p->read_so_far < len) {
+	g_at_chat_ref(p);
+
+	while (p->channel && (p->read_so_far < len)) {
 		gsize rbytes = MIN(len - p->read_so_far, wrap - p->read_so_far);
 		result = p->syntax->feed(p->syntax, (char *)buf, &rbytes);
 
@@ -628,8 +680,10 @@ static void new_bytes(GAtChat *p)
 	}
 
 	/* We're overflowing the buffer, shutdown the socket */
-	if (ring_buffer_avail(p->buf) == 0)
+	if (p->buf && ring_buffer_avail(p->buf) == 0)
 		g_source_remove(p->read_watch);
+
+	g_at_chat_unref(p);
 }
 
 static void debug_chat(GAtChat *chat, gboolean in, const char *str, gsize len)
@@ -753,23 +807,40 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 	return TRUE;
 }
 
+static void wakeup_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	GAtChat *chat = user_data;
+
+	if (ok == FALSE)
+		return;
+
+	if (chat->debugf)
+		chat->debugf("Finally woke up the modem\n", chat->debug_data);
+
+	g_source_remove(chat->timeout_source);
+	chat->timeout_source = 0;
+}
+
 static gboolean wakeup_no_response(gpointer user)
 {
 	GAtChat *chat = user;
 	struct at_command *cmd = g_queue_peek_head(chat->command_queue);
 
-	chat->timeout_source = 0;
-
-	/* Sometimes during startup the modem is still in the ready state
-	 * and might acknowledge our 'wakeup' command.  In that case don't
-	 * timeout the wrong command
-	 */
-	if (cmd == NULL || cmd->id != 0)
-		return FALSE;
+	if (chat->debugf)
+		chat->debugf("Wakeup got no response\n", chat->debug_data);
 
 	g_at_chat_finish_command(chat, FALSE, NULL);
+	cmd = at_command_create(chat->wakeup, none_prefix, FALSE,
+				NULL, wakeup_cb, chat, NULL, TRUE);
 
-	return FALSE;
+	if (!cmd) {
+		chat->timeout_source = 0;
+		return FALSE;
+	}
+
+	g_queue_push_head(chat->command_queue, cmd);
+
+	return TRUE;
 }
 
 static gboolean can_write_data(GIOChannel *channel, GIOCondition cond,
@@ -819,8 +890,8 @@ static gboolean can_write_data(GIOChannel *channel, GIOCondition cond,
 	}
 
 	if (chat->cmd_bytes_written == 0 && wakeup_first == TRUE) {
-		cmd = at_command_create(chat->wakeup, NULL, FALSE, NULL, NULL,
-					NULL, NULL);
+		cmd = at_command_create(chat->wakeup, none_prefix, FALSE,
+					NULL, wakeup_cb, chat, NULL, TRUE);
 
 		if (!cmd)
 			return FALSE;
@@ -990,7 +1061,16 @@ void g_at_chat_unref(GAtChat *chat)
 		return;
 
 	g_at_chat_shutdown(chat);
-	g_free(chat);
+
+	/* glib delays the destruction of the watcher until it exits, this
+	 * means we can't free the data just yet, even though we've been
+	 * destroyed already.  We have to wait until the read_watcher
+	 * destroy function gets called
+	 */
+	if (chat->read_watch != 0)
+		chat->destroyed = TRUE;
+	else
+		g_free(chat);
 }
 
 gboolean g_at_chat_shutdown(GAtChat *chat)
@@ -1058,7 +1138,7 @@ static guint send_common(GAtChat *chat, const char *cmd,
 		return 0;
 
 	c = at_command_create(cmd, prefix_list, expect_pdu, listing, func,
-				user_data, notify);
+				user_data, notify, FALSE);
 
 	if (!c)
 		return 0;
