@@ -158,8 +158,9 @@ static const char *phone_and_clip_to_string(const struct ofono_phone_number *n,
 static const char *time_to_str(const time_t *t)
 {
 	static char buf[128];
+	struct tm tm;
 
-	strftime(buf, 127, "%Y-%m-%dT%H:%M:%S%z", localtime(t));
+	strftime(buf, 127, "%Y-%m-%dT%H:%M:%S%z", localtime_r(t, &tm));
 	buf[127] = '\0';
 
 	return buf;
@@ -210,30 +211,6 @@ static DBusMessage *voicecall_get_properties(DBusConnection *conn,
 	return reply;
 }
 
-static DBusMessage *voicecall_busy(DBusConnection *conn,
-					DBusMessage *msg, void *data)
-{
-	struct voicecall *v = data;
-	struct ofono_voicecall *vc = v->vc;
-	struct ofono_call *call = v->call;
-
-	if (call->status != CALL_STATUS_INCOMING &&
-		call->status != CALL_STATUS_WAITING)
-		return __ofono_error_failed(msg);
-
-	if (!vc->driver->set_udub)
-		return __ofono_error_not_implemented(msg);
-
-	if (vc->pending)
-		return __ofono_error_busy(msg);
-
-	vc->pending = dbus_message_ref(msg);
-
-	vc->driver->set_udub(vc, generic_callback, vc);
-
-	return NULL;
-}
-
 static DBusMessage *voicecall_deflect(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -276,6 +253,7 @@ static DBusMessage *voicecall_hangup(DBusConnection *conn,
 	struct voicecall *v = data;
 	struct ofono_voicecall *vc = v->vc;
 	struct ofono_call *call = v->call;
+	int num_calls;
 
 	if (call->status == CALL_STATUS_DISCONNECTED)
 		return __ofono_error_failed(msg);
@@ -296,7 +274,19 @@ static DBusMessage *voicecall_hangup(DBusConnection *conn,
 		return NULL;
 	}
 
-	if ((g_slist_length(vc->call_list) == 1) && vc->driver->hangup &&
+	if (call->status == CALL_STATUS_WAITING) {
+		if (vc->driver->set_udub == NULL)
+			return __ofono_error_not_implemented(msg);
+
+		vc->pending = dbus_message_ref(msg);
+		vc->driver->set_udub(vc, generic_callback, vc);
+
+		return NULL;
+	}
+
+	num_calls = g_slist_length(vc->call_list);
+
+	if (num_calls == 1 && vc->driver->hangup &&
 			(call->status == CALL_STATUS_ACTIVE ||
 				call->status == CALL_STATUS_DIALING ||
 				call->status == CALL_STATUS_ALERTING)) {
@@ -306,6 +296,14 @@ static DBusMessage *voicecall_hangup(DBusConnection *conn,
 		return NULL;
 	}
 
+	if (num_calls == 1 && vc->driver->release_all_held &&
+			call->status == CALL_STATUS_HELD) {
+		vc->pending = dbus_message_ref(msg);
+		vc->driver->release_all_held(vc, generic_callback, vc);
+
+		return NULL;
+	}
+
 	if (vc->driver->release_specific == NULL)
 		return __ofono_error_not_implemented(msg);
 
@@ -341,8 +339,6 @@ static DBusMessage *voicecall_answer(DBusConnection *conn,
 
 static GDBusMethodTable voicecall_methods[] = {
 	{ "GetProperties",	"",	"a{sv}",	voicecall_get_properties },
-	{ "Busy",		"",	"",		voicecall_busy,
-							G_DBUS_METHOD_FLAG_ASYNC },
 	{ "Deflect",		"s",	"",		voicecall_deflect,
 							G_DBUS_METHOD_FLAG_ASYNC },
 	{ "Hangup",		"",	"",		voicecall_hangup,
@@ -579,7 +575,7 @@ static gboolean voicecalls_can_dtmf(struct ofono_voicecall *vc)
 	return FALSE;
 }
 
-static gboolean voicecalls_have_held(struct ofono_voicecall *vc)
+static gboolean voicecalls_have_with_status(struct ofono_voicecall *vc, int status)
 {
 	GSList *l;
 	struct voicecall *v;
@@ -587,13 +583,18 @@ static gboolean voicecalls_have_held(struct ofono_voicecall *vc)
 	for (l = vc->call_list; l; l = l->next) {
 		v = l->data;
 
-		if (v->call->status == CALL_STATUS_HELD)
+		if (v->call->status == status)
 			return TRUE;
 	}
 
 	return FALSE;
 }
 
+static gboolean voicecalls_have_held(struct ofono_voicecall *vc)
+{
+	return voicecalls_have_with_status(vc, CALL_STATUS_HELD);
+}
+
 static int voicecalls_num_with_status(struct ofono_voicecall *vc,
 					int status)
 {
@@ -673,17 +674,12 @@ static GSList *voicecalls_active_list(struct ofono_voicecall *vc)
 
 static gboolean voicecalls_have_waiting(struct ofono_voicecall *vc)
 {
-	GSList *l;
-	struct voicecall *v;
+	return voicecalls_have_with_status(vc, CALL_STATUS_WAITING);
+}
 
-	for (l = vc->call_list; l; l = l->next) {
-		v = l->data;
-
-		if (v->call->status == CALL_STATUS_WAITING)
-			return TRUE;
-	}
-
-	return FALSE;
+static gboolean voicecalls_have_incoming(struct ofono_voicecall *vc)
+{
+	return voicecalls_have_with_status(vc, CALL_STATUS_INCOMING);
 }
 
 static gboolean real_emit_call_list_changed(void *data)
@@ -950,33 +946,6 @@ static void dial_callback(const struct ofono_error *error, void *data)
 		emit_call_list_changed(vc);
 }
 
-static void swap_before_dial(const struct ofono_error *error, void *data)
-{
-	struct ofono_voicecall *vc = data;
-	const char *number;
-	const char *clirstr;
-	struct ofono_phone_number ph;
-	enum ofono_clir_option clir;
-
-	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Couldn't swap for dialing because: %s",
-				telephony_error_to_str(error));
-		__ofono_dbus_pending_reply(&vc->pending,
-					__ofono_error_failed(vc->pending));
-		return;
-	}
-
-	dbus_message_get_args(vc->pending, NULL, DBUS_TYPE_STRING, &number,
-					DBUS_TYPE_STRING, &clirstr,
-					DBUS_TYPE_INVALID);
-
-	string_to_phone_number(number, &ph);
-	clir_string_to_clir(clirstr, &clir);
-
-	vc->driver->dial(vc, &ph, clir, OFONO_CUG_OPTION_DEFAULT,
-				dial_callback, vc);
-}
-
 static DBusMessage *manager_dial(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -985,7 +954,6 @@ static DBusMessage *manager_dial(DBusConnection *conn,
 	struct ofono_phone_number ph;
 	const char *clirstr;
 	enum ofono_clir_option clir;
-	gboolean have_active;
 
 	if (vc->pending)
 		return __ofono_error_busy(msg);
@@ -1007,44 +975,15 @@ static DBusMessage *manager_dial(DBusConnection *conn,
 	if (!vc->driver->dial)
 		return __ofono_error_not_implemented(msg);
 
-	have_active = voicecalls_have_active(vc);
-
-	if (have_active && voicecalls_have_held(vc))
+	if (voicecalls_have_incoming(vc))
 		return __ofono_error_failed(msg);
 
-	/* There are three cases we need to watch out for:
-	 * - No Active Calls - Just Dial
-	 * - Active Call - We have to swap first
-	 * - Active and Waiting call - If we can swap without accepting,
-	 *   	then do so, otherwise we should fail.
-	 */
+	/* We can't have two dialing/alerting calls, reject outright */
+	if (voicecalls_num_connecting(vc) > 0)
+		return __ofono_error_failed(msg);
 
-	if (have_active && voicecalls_have_waiting(vc)) {
-		if (vc->driver->swap_without_accept == NULL)
-			return __ofono_error_failed(msg);
-
-		vc->pending = dbus_message_ref(msg);
-
-		vc->driver->swap_without_accept(vc, swap_before_dial, vc);
-
-		return NULL;
-	}
-
-	if (have_active) {
-		if (vc->driver->swap_without_accept == NULL &&
-				vc->driver->hold_all_active == NULL)
-			return __ofono_error_failed(msg);
-
-		vc->pending = dbus_message_ref(msg);
-
-		if (vc->driver->swap_without_accept)
-			vc->driver->swap_without_accept(vc, swap_before_dial,
-							vc);
-		else
-			vc->driver->hold_all_active(vc, swap_before_dial, vc);
-
-		return NULL;
-	}
+	if (voicecalls_have_active(vc) && voicecalls_have_held(vc))
+		return __ofono_error_failed(msg);
 
 	vc->pending = dbus_message_ref(msg);
 
@@ -1534,7 +1473,7 @@ void ofono_voicecall_disconnected(struct ofono_voicecall *vc, int id,
 
 	ofono_debug("Got disconnection event for id: %d, reason: %d", id, reason);
 
-	l = g_slist_find_custom(vc->call_list, GINT_TO_POINTER(id),
+	l = g_slist_find_custom(vc->call_list, GUINT_TO_POINTER(id),
 				call_compare_by_id);
 
 	if (!l) {
@@ -1548,7 +1487,7 @@ void ofono_voicecall_disconnected(struct ofono_voicecall *vc, int id,
 	ts = time(NULL);
 	prev_status = call->call->status;
 
-	l = g_slist_find_custom(vc->multiparty_list, GINT_TO_POINTER(id),
+	l = g_slist_find_custom(vc->multiparty_list, GUINT_TO_POINTER(id),
 				call_compare_by_id);
 
 	if (l) {
@@ -1592,12 +1531,12 @@ void ofono_voicecall_notify(struct ofono_voicecall *vc,
 	struct ofono_modem *modem = __ofono_atom_get_modem(vc->atom);
 	GSList *l;
 	struct voicecall *v = NULL;
-	struct ofono_call *newcall = NULL;
+	struct ofono_call *newcall;
 
 	ofono_debug("Got a voicecall event, status: %d, id: %u, number: %s",
 			call->status, call->id, call->phone_number.number);
 
-	l = g_slist_find_custom(vc->call_list, GINT_TO_POINTER(call->id),
+	l = g_slist_find_custom(vc->call_list, GUINT_TO_POINTER(call->id),
 				call_compare_by_id);
 
 	if (l) {
@@ -1611,15 +1550,13 @@ void ofono_voicecall_notify(struct ofono_voicecall *vc,
 
 	ofono_debug("Did not find a call with id: %d\n", call->id);
 
-	newcall = g_try_new0(struct ofono_call, 1);
+	newcall = g_memdup(call, sizeof(struct ofono_call));
 
-	if (!call) {
+	if (!newcall) {
 		ofono_error("Unable to allocate call");
 		goto err;
 	}
 
-	memcpy(newcall, call, sizeof(struct ofono_call));
-
 	if (__ofono_modem_alloc_callid(modem) != call->id) {
 		ofono_error("Warning: Call id and internally tracked id"
 				" do not correspond");
